# Copyright (c) 2003-2004 Chalmers University of Technology
# Authors: Jean-Marc Orliaguet <jmo@ita.chalmers.se>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 02111-1307, USA.
#

__author__ = "Jean-Marc Orliaguet <jmo@ita.chalmers.se>"

"""
  Page Block
"""

from Globals import InitializeClass
from AccessControl import ClassSecurityInfo
from OFS.Folder import Folder

from Products.CMFCore.CMFCorePermissions import View
from PortalContentFolder import PortalContentFolder

from cpsskins_utils import rebuild_properties, callAction, \
                           getFreeId, verifyThemePerms

def addPageBlock(dispatcher, id, REQUEST=None, **kw):
    """Add a Page Block."""
    ob = PageBlock(id)
    container = dispatcher.Destination()
    container._setObject(id, ob)
    if REQUEST is not None:
        url = dispatcher.DestinationURL()
        REQUEST.RESPONSE.redirect('%s/manage_main' % url)


factory_type_information = (
    {'id': 'Page Block',
     'title': 'Page Block',
     'description': ('_pageblock_description_'),
     'meta_type': 'Page Block',
     'icon': 'pageblock.gif',
     'product': 'CPSSkins',
     'factory': 'addPageBlock',
     'immediate_view': 'cpsskins_edit_form',
     'filter_content_types': 0,
     'global_allow': 0,
     'actions': (
                 {'id': 'view',
                  'name': 'View',
                  'action': 'cpsskins_default_view',
                  'permissions': (View,)},
                 {'id': 'edit',
                  'name': 'Edit',
                  'action': 'cpsskins_edit_form',
                  'permissions': ('Manage Themes',)},
                 {'id': 'delete',
                  'name': 'Delete',
                  'action': 'cpsskins_object_delete',
                  'visible': 0,
                  'permissions': ('Manage Themes',)},
                 {'id': 'moveup',
                  'name': 'Move Up',
                  'action': 'cpsskins_pageblock_moveup',
                  'visible': 0,
                  'permissions': ('Manage Themes',)},
                 {'id': 'movedown',
                  'name': 'Move Down',
                  'action': 'cpsskins_pageblock_movedown',
                  'visible': 0,
                  'permissions': ('Manage Themes',)},
                 {'id': 'toggle',
                  'name': 'Toggle',
                  'action': 'cpsskins_object_toggle',
                  'visible': 0,
                  'permissions': ('Manage Themes',)},
                 {'id': 'addtemplet',
                  'name': '_action_addtemplet_',
                  'action': 'cpsskins_addtemplet_form',
                  'visible': 0,
                  'permissions': ('Manage Themes',),
                  'category': 'object'
                  },
                 ),
     },
    )


class PageBlock(PortalContentFolder):
    """
    Class for page blocks.
    """

    meta_type = "Page Block"
    portal_type = "Page Block"

    manage_options = ( 
                      Folder.manage_options
                     )

    security = ClassSecurityInfo()

    _properties = (
        {'id': 'title', 
         'type': 'string', 
         'mode': 'w', 
         'label': 'Title', 
         'category': 'general'
        },
        {'id': 'closed', 
         'type': 'boolean', 
         'mode': 'w', 
         'label': 'Closed', 
         'category': 'none',
         'default': 0
        },
        {'id': 'width', 
         'type': 'string', 
         'mode': 'w', 
         'label': 'Width', 
         'category': 'general' 
        },
        {'id': 'height', 
         'type': 'string', 
         'mode': 'w', 
         'label': 'Height', 
         'category': 'general'
        },
        {'id': 'maxcols', 
         'type': 'int', 
         'mode': 'w', 
         'label': 'Number of columns', 
         'category': 'general'
        },
        {'id': 'shape', 
         'type': 'selection', 
         'mode': 'w', 
         'label': 'Shape', 
         'select_variable': 'cpsskins_listAreaShapes', 
         'meta_type': 'Area Shape', 
         'category' : 'general'
        },
        {'id': 'color', 
         'type': 'selection', 
         'mode': 'w', 
         'label': 'Color', 
         'select_variable': 'cpsskins_listAreaColors', 
         'meta_type': 'Area Color', 
         'category' : 'general'
        },
      )

    # XXX to be fixed: title
    def __init__(self, id, 
                       title = '',
                       closed = 0,
                       maxcols = 1,
                       width = '100%', 
                       height = '', 
                       shape = 'NoBorder', 
                       color = 'Transparent', 
                       **kw):

        self.id = id
        self.title = title
        self.closed = closed
        self.maxcols = maxcols
        self.width = width
        self.height = height
        self.shape = shape
        self.color = color

    security.declareProtected(View,'isPortalPageBlock')
    def isPortalPageBlock(self):
        """ is Portal Page Block ? """   
        return 1

    security.declarePublic('getIconRelative')
    def getIconRelative(self):
        """
        Gets the Icon path, relative to the portal.
        This is needed to catalog correctly in the presence of VHM.
        """
        return self.getIcon(1)

    security.declareProtected('Manage Themes', 'rebuild')
    def rebuild(self):
        """                  
        Rebuild this page block
        """              
                         
        rebuild_properties(self)
        verifyThemePerms(self)

        for obj in self.objectValues():
           meth = getattr(obj.aq_explicit, 'isPortalTemplet', None)
           if callable(meth) and meth():
                  obj.rebuild()
           meth = getattr(obj.aq_explicit, 'isCellModifier', None)
           if callable(meth) and meth():
                  obj.rebuild()


    security.declarePrivate('getActions')
    def getActions(self):

        return self.portal_actions.listFilteredActionsFor(self)


    security.declarePublic('getInfo')
    def getInfo(self):

        infoblock = {}
        actions_list = ['moveup', 'movedown', 'toggle', 'delete']

        ti = self.getTypeInfo()
        for actionid in actions_list:  
             actioninfo = {}
             if actionid == 'moveup':
                 actioninfo['can_move'] = self.can_moveup()
             if actionid == 'movedown':
                 actioninfo['can_move'] = self.can_movedown()
             if actionid == 'toggle':
                 actioninfo['can_toggle'] = self.can_toggle()
             if actionid == 'delete':
                 actioninfo['can_delete'] = self.can_delete()
             try:
                action = ti.getActionById(actionid)
                actioninfo['url']  = self.absolute_url() + '/' + self.restrictedTraverse(action).getId()
             except:
                continue
             infoblock[actionid] = actioninfo

        return infoblock


    security.declareProtected('Manage Themes', 'add_templet_form')
    def add_templet_form(self, **kw):
        """
        Call the add templet action.
        """
        return callAction(self, 'addtemplet', **kw)


    security.declareProtected('Manage Themes', 'addTemplet')
    def addTemplet(self, **kw):
        """
        Add a Templet. Returns the Templet's id
        """

        type_name = kw.get('type_name', None)
        if type_name is None:
            return
        del kw['type_name']

        xpos = kw.get('xpos', 0)
        ypos = kw.get('ypos', 0)

        kw['title'] = type_name
        ypos = int(ypos)
        if ypos == 0:
            newpos = ypos 
        else:
            newpos = ypos + 1

        id = getFreeId(self)
        self.invokeFactory(type_name, id, **kw)
        templet = getattr(self, id, None) 
        if templet is not None:
            self.move_object_to_position(templet.getId(), newpos)
            templet.xpos = int(xpos)
            return templet
        else:
            return


    security.declareProtected('Manage Themes', 'addCellHider')
    def addCellHider(self, **kw):
        """
        Add a Cell Hider. Returns the Cell Hider's id.
        """

        id = getFreeId(self)

        cellhider_xpos = kw.get('cellhider_xpos', None)
        if cellhider_xpos is not None:
            self.invokeFactory('Cell Hider', id, xpos=int(cellhider_xpos))
            cellhider = getattr(self, id, None) 
            return cellhider
        else:
            return

    security.declareProtected('Manage Themes', 'addCellSizer')
    def addCellSizer(self, **kw):
        """
        Add a Cell Sizer. Returns the Cell Sizer's id.
        """

        id = getFreeId(self)

        cellsizer_xpos = kw.get('cellsizer_xpos', None)
        if cellsizer_xpos is not None:
            self.invokeFactory('Cell Sizer', id, xpos=int(cellsizer_xpos))
            cellsizer = getattr(self, id, None) 
            return cellsizer
        else:
            return

    security.declareProtected('Manage Themes', 'addCellStyler')
    def addCellStyler(self, **kw):
        """
        Add a Cell Styler. Returns the Cell Styler's id.
        """

        id = getFreeId(self)

        cellstyler_xpos = kw.get('cellstyler_xpos', None)
        if cellstyler_xpos is not None:
            self.invokeFactory('Cell Styler', id, xpos=int(cellstyler_xpos))
            cellstyler = getattr(self, id, None) 
            return cellstyler
        else:
            return


    security.declareProtected('Manage Themes', 'moveCell')
    def moveCell(self, **kw):
        """
        Move a cell into a given direction.
        """

        xpos = kw.get('xpos', None)
        dir = kw.get('dir', None)

        if xpos is None or dir not in ['left', 'right']:
            return 
        
        delta = 0
        if dir == 'left' and int(xpos) > 0:
            delta = -1
        
        if dir == 'right' and int(xpos) <= int(self.maxcols):
            delta = +1

        if delta == 0:
            return

        objects_src = []
        for obj in self.objectValues():
            if hasattr(obj, 'xpos') and int(obj.xpos) == int(xpos):
                objects_src.append(obj)

        objects_dest = []
        for obj in self.objectValues():
            if hasattr(obj, 'xpos') and int(obj.xpos) == int(xpos) + delta:
                objects_dest.append(obj)


        for obj in objects_src:
            obj.xpos = obj.xpos + delta

        for obj in objects_dest:
            obj.xpos = obj.xpos - delta



    security.declarePublic('getObjects')
    def getObjects(self):
        """
        Gets the templets of a user in a page block 
        """
        
        templet_dict = {}
        cellsizer_dict = {}
        cellstyler_dict = {}
        cellhider_dict = {}

        for objid, obj in self.objectItems():
            meth = getattr(obj, 'isPortalTemplet', None)
            if callable(meth) and meth():
                 templet_dict.setdefault(obj.xpos, []).append(obj)

            meth = getattr(obj, 'isCellStyler', None)
            if callable(meth) and meth():
                      cellstyler_dict[obj.xpos] = obj

            meth = getattr(obj, 'isCellSizer', None)
            if callable(meth) and meth():
                      cellsizer_dict[obj.xpos] = obj

            meth = getattr(obj, 'isCellHider', None)
            if callable(meth) and meth():
                      cellhider_dict[obj.xpos] = obj

        main_dict = {}
        main_dict['templets'] = templet_dict
        main_dict['cellstylers'] = cellstyler_dict
        main_dict['cellsizers'] = cellsizer_dict
        main_dict['cellhiders'] = cellhider_dict

        return main_dict

    security.declareProtected(View, 'can_toggle')
    def can_toggle(self):
        """
        Can the page block be toggled ( open/close ) ?
        """

        return self.can_delete()

    security.declareProtected(View, 'can_delete')
    def can_delete(self):
        """ Can the page block be deleted ?
            Not if it contains a 'main content' templet
        """
           
        for obj in self.objectValues():
                  meth = getattr(obj, 'isMainContent', None)
                  if meth is not None and meth(): 
                      return 0
        return 1


    security.declareProtected('Manage Themes', 'toggle')
    def toggle(self, state):
        """
        Open / close the page block
        """

        self.closed = state

        return 

    security.declareProtected('Manage Themes', 'getPageBlockPosition')
    def getPageBlockPosition(self):
        """ 
        Return the page block's ypos in the theme folder
        """

        container = self.aq_parent
        return container.get_object_position(self.getId())


    security.declareProtected('Manage Themes', 'move')
    def move(self, direction=None):
        """
        Move the page block into a direction 
        """
    
        container = self.aq_parent

        if direction == 'up' and self.can_moveup():
            newpos = self.moveup_pos()
            container.move_object_to_position(self.getId(), newpos)
    
        if direction == 'down' and self.can_movedown():
            newpos = self.movedown_pos()
            container.move_object_to_position(self.getId(), newpos)

        return self

    security.declareProtected(View, 'can_moveup')
    def can_moveup(self):
        """
        Can the page block be moved up?
        """

        container = self.aq_parent
        this_pos = container.get_object_position(self.getId())
        for obj in container.objectValues():
               meth = getattr(obj, 'isPortalPageBlock', None)
               if callable(meth) and meth():
                  pos = container.get_object_position(obj.getId())
                  if pos < this_pos:
                     return 1
        return

    security.declareProtected(View, 'can_movedown')
    def can_movedown(self):
        """
        Can the page block be moved down?
        """

        container = self.aq_parent
        this_pos = container.get_object_position(self.getId())
        for obj in container.objectValues():
               meth = getattr(obj, 'isPortalPageBlock', None)
               if callable(meth) and meth():
                  pos = container.get_object_position(obj.getId())
                  if pos > this_pos:
                     return 1
        return

    security.declareProtected(View, 'moveup_pos')
    def moveup_pos(self):
        """
        Return the new position 
        """

        container = self.aq_parent
        this_pos = container.get_object_position(self.getId())
        newpos = -1
        for obj in container.objectValues():
            meth = getattr(obj, 'isPortalPageBlock', None)
            if callable(meth) and meth():
               pos = container.get_object_position(obj.getId())
               if pos > newpos and pos < this_pos:
                  newpos = pos

        if newpos >= 0:
           return newpos

    security.declareProtected(View, 'movedown_pos')
    def movedown_pos(self):
        """
        Return the new position 
        """
          
        container = self.aq_parent
        this_pos = container.get_object_position(self.getId())
        for obj in container.objectValues():
            meth = getattr(obj, 'isPortalPageBlock', None)
            if callable(meth) and meth():
               pos = container.get_object_position(obj.getId())
               if  pos > this_pos:
                  return pos      
                                    
        return    

    security.declareProtected('Manage Themes', 'edit_form')
    def edit_form(self, **kw):
        """
        Call the edit action.
        """
        return callAction(self, 'edit', **kw)


    security.declareProtected('Manage Themes', 'edit')
    def edit(self, **kw):
        """
        Default edit method, changes the properties.
        """
        self.manage_changeProperties(**kw)


InitializeClass(PageBlock)


